/*
 * $Id: ft_push.c,v 1.1 2003/07/06 13:05:04 jasta Exp $
 *
 * Copyright (C) 2001-2003 giFT project (gift.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "ft_openft.h"

#include "ft_transfer.h"
#include "ft_http_client.h"

#include "ft_protocol.h"

#include "ft_push.h"

/*****************************************************************************/

FT_HANDLER (ft_push_request)
{
	in_addr_t  ip;
	in_port_t  port;
	char      *request;

	ip      = ft_packet_get_ip     (packet);
	port    = ft_packet_get_uint16 (packet, TRUE);
	request = ft_packet_get_str    (packet);

	/* sanity check on the data received */
	if (!request)
		return;

	/* ip || port == 0 in this instance means that we are to send this file
	 * back to the search node contacting us here */
	if (ip == 0 || port == 0)
	{
		ip   = FT_NODE(c)->ip;
		port = FT_NODE(c)->http_port;
	}

	/* make an outgoing http connection and advertise that we can fulfill
	 * the push request */
	ft_http_client_push (ip, port, request);
}

FT_HANDLER (ft_push_forward)
{
	FTPacket *fwd;
	in_addr_t ip;
	char     *request;

	ip      = ft_packet_get_ip  (packet);
	request = ft_packet_get_str (packet);

	if (ip == 0 || !request)
	{
		FT->DBGSOCK (FT, c, "invalid push request");
		return;
	}

	if (!(fwd = ft_packet_new (FT_PUSH_REQUEST, 0)))
		return;

	/* reconstruct the complete request using the directly connected peers
	 * address information */
	ft_packet_put_ip     (fwd, FT_NODE(c)->ip);
	ft_packet_put_uint16 (fwd, FT_NODE(c)->http_port, TRUE);

	/* put the original data back on */
	ft_packet_put_str    (fwd, request);

	/*
	 * Pass this data along to the user who will be actually sending the
	 * file to the user making this request.
	 *
	 * NOTE: No further relayed communication will take place.  If
	 * successful, the node will connect back to the user and advertise the
	 * file, otherwise it will just be dropped off into oblivion right here.
	 * Life is rough, eh? :)
	 */
	ft_packet_sendto (ip, fwd);
}
